#!/bin/sh
# build.sh -- 
# Created: Mon Jul 14 10:53:31 1997 by faith@acm.org
# Revised: Mon Jul 14 16:22:18 1997 by faith@acm.org
# Copyright 1997 Rickard E. Faith (faith@acm.org)
# This program comes with ABSOLUTELY NO WARRANTY.
# 
# $Id: build.sh,v 1.1 1997/07/19 05:35:21 faith Exp $
# 
#

ARCHIVE=gazetteer

zcat $ARCHIVE/places.zip | ./a.out -p > tmp
zcat $ARCHIVE/counties.zip | ./a.out -c >> tmp
# zcat $ARCHIVE/mcds.zip | ./a.out -m >> tmp
zcat $ARCHIVE/zips.zip | ./a.out -z >> tmp

sort tmp > tmp.sort

